==> openapi_tools_abm-5.0.65180a0.tar.gz: openapi_tools_abm-5.0.65180a0/openapi_tools/open_api_server.py <==
#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors:
#  Mike Orlov <m.orlov@techokert.ru>
import dataclasses
import json
import logging
import traceback
import typing
import warnings
from dataclasses import dataclass, field, MISSING
from functools import cached_property
from typing import Callable, Any, Awaitable, TypeVar

import http_tools
from async_tools import acall
from dict_caster.extras import first
from http_tools import Answer, HttpServer, HttpStatusCode
from http_tools.answer import ExceptionAnswer, ErrorDescription, JsonableAnswer, FileAnswer
from http_tools.mime_types import ContentType
from http_tools.request import IncomingRequest
from init_helpers import custom_dumps
from init_helpers.dict_to_dataclass import NoValue

from .open_api_wrapper import OpenApiWrapper, Endpoint, RequestBody, Content
from .parameter import CallParameter, SpecParameter, RawBodyParameter, ParameterAggregation, BodyParameter
from .security import Unauthorized, Forbidden, Security, get_securities_arg_names, resolver_cache, SecurityScheme

logger = logging.getLogger(__name__)
ResultType = TypeVar('ResultType', bound=Any)
AnswerType = TypeVar('AnswerType', bound=Answer)


def _return_empty_dict(*_: Any, **__: Any):
    return {}


@dataclass(frozen=True)
class RpcEndpoint(typing.Generic[ResultType, AnswerType]):
    func: Callable[..., ResultType]
    method: str
    path: str
    securities: list[Security]
    arg_name_to_param: dict[str, CallParameter]
    answer_type: type[AnswerType]
    exception_type_to_answer_type: dict[type[Exception], type[ExceptionAnswer]]
    unhandled_exception__answer_type: type[Answer]
    operation_id: str
    answer_factory: Callable[[ResultType], AnswerType]

    def __init__(
            self,
            func: Callable[..., ResultType],
            method: str,
            path: str,
            securities: list[Security | SecurityScheme],
            arg_name_to_param: dict[str, CallParameter | type[RawBodyParameter]],
            answer_type: type[AnswerType],
            exception_type_to_answer_type: dict[type[Exception], type[ExceptionAnswer]],
            unhandled_exception__answer_type: type[Answer] = ExceptionAnswer[HttpStatusCode.InternalServerError],
            operation_id: str = '',
            answer_factory: Callable[[ResultType], AnswerType] | None = None,
    ) -> None:
        object.__setattr__(self, 'func', func)
        object.__setattr__(self, 'method', method)
        object.__setattr__(self, 'path', path)
        object.__setattr__(self, 'securities', [s if isinstance(s, Security) else s.has() for s in securities])
        object.__setattr__(self, 'arg_name_to_param', {
            a: p() if isinstance(p, type) else p for a, p in arg_name_to_param.items()})
        object.__setattr__(self, 'answer_type', answer_type)
        object.__setattr__(self, 'exception_type_to_answer_type', exception_type_to_answer_type)
        object.__setattr__(self, 'unhandled_exception__answer_type', unhandled_exception__answer_type)
        object.__setattr__(self, 'operation_id', operation_id or self.func.__name__)
        object.__setattr__(self, 'answer_factory', answer_factory or answer_type)
        get_securities_arg_names(self.securities)

    @cached_property
    def parameters(self) -> set[CallParameter]:
        result = set()
        for parameter in self.arg_name_to_param.values():
            if isinstance(parameter, ParameterAggregation):
                result.update(parameter.get_parameters())
            elif isinstance(parameter, CallParameter):
                result.add(parameter)
            else:
                raise TypeError(f"Unexpected parameter type: {type(parameter)}")
        return result

    @cached_property
    def body_parameters(self) -> list[BodyParameter]:
        body_params: list[BodyParameter] = [param for param in self.parameters if isinstance(param, BodyParameter)]
        assert len(param_types := {type(param) for param in body_params}) <= 1, f"Can not combine {param_types} in body"
        assert len(param_mimes := {param.body_mime_type for param in body_params}) <= 1, \
            f"Can not combine {param_mimes} in body"
        return sorted(body_params, key=lambda x: x.default is not NoValue)

    @cached_property
    def body_parameter_type(self) -> type[BodyParameter] | None:
        return type(first(self.body_parameters, none_if_empty=True))

    @cached_property
    def non_body_parameters(self) -> list[CallParameter]:
        return [param for param in self.parameters if not isinstance(param, BodyParameter)]

    @cached_property
    def body_content_schema(self) -> type | None:
        if not (body_params := self.body_parameters):
            return None

        if self.body_parameter_type and issubclass(self.body_parameter_type, RawBodyParameter):
            return bytes

        class_: type = type(self.operation_id, tuple(), {
            p.name: field(default=MISSING if p.default is not NoValue else p.default) for p in body_params})
        class_.__annotations__ = {param.name: param.schema for param in body_params}
        # noinspection PyTypeChecker
        return dataclasses.dataclass(class_)

    @cached_property
    def request_body(self) -> RequestBody | None:
        if (body_parameter_type := self.body_parameter_type) and (body_content_schema := self.body_content_schema):
            return RequestBody(Content(body_parameter_type.body_mime_type, body_content_schema))

    @cached_property
    def code_to_answer(self) -> dict[http_tools.HttpStatusCode, type[http_tools.Answer]]:
        result = {self.answer_type.get_class_status_code(): self.answer_type}
        result |= {ans_t.get_class_status_code(): ans_t for ans_t in self.exception_type_to_answer_type.values()}
        if self.unhandled_exception__answer_type is not None:
            unhandled_exception__answer_code = self.unhandled_exception__answer_type.get_class_status_code()
            result[unhandled_exception__answer_code] = self.unhandled_exception__answer_type
        return result

    def as_endpoint(self) -> Endpoint:
        return Endpoint(
            path=self.path,
            method=self.method,
            operation_id=self.operation_id,
            securities=self.securities,
            parameters=[param for param in self.non_body_parameters if isinstance(param, SpecParameter)],
            request_body=self.request_body,
            code_to_answer=self.code_to_answer
        )

    def gen_handler(self) -> Callable[[IncomingRequest], Awaitable[Answer]]:
        # TODO: check argument fullness
        if not self.securities:
            warnings.warn('Endpoints without securities will be prohibited')

        async def handler(request: IncomingRequest) -> Answer:
            kwargs = None
            cache_token = resolver_cache.set({})
            security_key_to_exception = {}
            if self.securities:
                for security in self.securities:
                    security_key = f'{security.scheme.key}[{",".join(security.scopes)}]'
                    try:
                        kwargs = await security.evaluate(request)
                        break
                    except (Unauthorized, Forbidden) as e:
                        if security.scheme.do_log:
                            security_key_to_exception[security_key] = e
            else:
                kwargs = {}
            resolver_cache.reset(cache_token)

            if kwargs is None:
                is_forbidden = any(isinstance(exc, Forbidden) for exc in security_key_to_exception.values())
                status_code = HttpStatusCode.Forbidden if is_forbidden else HttpStatusCode.Unauthorized
                # noinspection PyTypeChecker
                return JsonableAnswer(ErrorDescription(security_key_to_exception, "ErrorGroup"), status_code)

            try:
                kwargs |= {name: param.get(request) for name, param in self.arg_name_to_param.items()}
                call_result = await acall(self.func(**kwargs))
                answer_factory = self.answer_type if self.answer_factory is None else self.answer_factory
                return answer_factory(call_result)
            except Exception as e:
                for exception_type, answer_type in self.exception_type_to_answer_type.items():
                    if isinstance(e, exception_type):
                        return answer_type(e)

                if self.unhandled_exception__answer_type is not None:
                    traceback.print_exc()
                    logger.exception(e)
                    return self.unhandled_exception__answer_type(e)

                raise

        return handler


class OpenApiServer:
    @dataclass
    class Context(OpenApiWrapper.Context):
        http_server: HttpServer

    @dataclass
    class Config(OpenApiWrapper.Config):
        write_spec_to_file: str = ""  # empty means skip
        spec_http_path: str = '/open_api_specification/get'  # empty means disable

    def __init__(
            self, config: Config, context: Context,
            wrapper_factory: Callable[[Config, Context], OpenApiWrapper] = OpenApiWrapper
    ):
        self.context = context
        self.config = config
        self.wrapper = wrapper_factory(config, context)
        if path := self.config.spec_http_path:
            self.context.http_server.register_handler(path, self.get_specification)

    async def get_specification(self, _: IncomingRequest) -> FileAnswer:
        content = custom_dumps(self.wrapper.get_spec_dict(), indent=2)
        return FileAnswer(content, file_name='openapi_spec.json', content_type=ContentType.Json)

    def _write_spec_to_file(self):
        if self.config.write_spec_to_file:
            spec = self.wrapper.get_spec_dict()
            json_res = json.dumps(spec, indent=2)
            with open(self.config.write_spec_to_file, "w") as f:
                f.write(json_res)

    def register_endpoint(self, rpc_endpoint: RpcEndpoint) -> None:
        self.context.http_server.register_handler(rpc_endpoint.path, rpc_endpoint.gen_handler(), [rpc_endpoint.method])
        self.wrapper.register_endpoint(rpc_endpoint.as_endpoint())
        self._write_spec_to_file()

==> openapi_tools_abm-5.0.65196a0.tar.gz: openapi_tools_abm-5.0.65196a0/openapi_tools/open_api_server.py <==
#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors:
#  Mike Orlov <m.orlov@techokert.ru>
import dataclasses
import json
import logging
import traceback
import typing
import warnings
from dataclasses import dataclass, field, MISSING
from functools import cached_property
from typing import Callable, Any, Awaitable, TypeVar

import http_tools
from async_tools import acall
from dict_caster.extras import first
from http_tools import Answer, HttpServer, HttpStatusCode
from http_tools.answer import ExceptionAnswer, ErrorDescription, JsonableAnswer, FileAnswer
from http_tools.mime_types import ContentType
from http_tools.request import IncomingRequest
from init_helpers import custom_dumps
from init_helpers.dict_to_dataclass import NoValue

from .open_api_wrapper import OpenApiWrapper, Endpoint, RequestBody, Content
from .parameter import CallParameter, SpecParameter, RawBodyParameter, ParameterAggregation, BodyParameter
from .security import Unauthorized, Forbidden, Security, get_securities_arg_names, resolver_cache, SecurityScheme

logger = logging.getLogger(__name__)
ResultType = TypeVar('ResultType', bound=Any)
AnswerType = TypeVar('AnswerType', bound=Answer)


def _return_empty_dict(*_: Any, **__: Any):
    return {}


@dataclass(frozen=True)
class RpcEndpoint(typing.Generic[ResultType, AnswerType]):
    func: Callable[..., ResultType]  # обработчик
    method: str   # HTTP метод
    path: str  # HTTP путь
    securities: list[Security]  # способы проверки авторизации, при пустом массиве проверка не выполняется
    arg_name_to_param: dict[str, CallParameter]  # описание получения значений для входных аргументов обработчика
    answer_type: type[AnswerType]  # Тип HTTP ответа
    # отображение ожидаемых исключений обработчика на типы HTTP ответов
    exception_type_to_answer_type: dict[type[Exception], type[ExceptionAnswer]]
    # тип HTTP ответа при возникновении неожиданного исключения
    unhandled_exception__answer_type: type[Answer]
    # идентификатор эндпоинта
    operation_id: str
    # преобразователь ответа обработчика в HTTP ответ
    answer_factory: Callable[[ResultType], AnswerType]

    def __init__(
            self,
            func: Callable[..., ResultType],
            method: str,
            path: str,
            securities: list[Security | SecurityScheme],
            arg_name_to_param: dict[str, CallParameter | type[RawBodyParameter]],
            answer_type: type[AnswerType],
            exception_type_to_answer_type: dict[type[Exception], type[ExceptionAnswer]] | None = None,
            unhandled_exception__answer_type: type[Answer] = ExceptionAnswer[HttpStatusCode.InternalServerError],
            operation_id: str = '',
            answer_factory: Callable[[ResultType], AnswerType] | None = None,
    ) -> None:
        object.__setattr__(self, 'func', func)
        object.__setattr__(self, 'method', method)
        object.__setattr__(self, 'path', path)
        object.__setattr__(self, 'securities', [s if isinstance(s, Security) else s.has() for s in securities])
        object.__setattr__(self, 'arg_name_to_param', {
            a: p() if isinstance(p, type) else p for a, p in arg_name_to_param.items()})
        object.__setattr__(self, 'answer_type', answer_type)
        object.__setattr__(self, 'exception_type_to_answer_type', exception_type_to_answer_type or {})
        object.__setattr__(self, 'unhandled_exception__answer_type', unhandled_exception__answer_type)
        object.__setattr__(self, 'operation_id', operation_id or self.func.__name__)
        object.__setattr__(self, 'answer_factory', answer_factory or answer_type)
        get_securities_arg_names(self.securities)

    @cached_property
    def parameters(self) -> set[CallParameter]:
        result = set()
        for parameter in self.arg_name_to_param.values():
            if isinstance(parameter, ParameterAggregation):
                result.update(parameter.get_parameters())
            elif isinstance(parameter, CallParameter):
                result.add(parameter)
            else:
                raise TypeError(f"Unexpected parameter type: {type(parameter)}")
        return result

    @cached_property
    def body_parameters(self) -> list[BodyParameter]:
        body_params: list[BodyParameter] = [param for param in self.parameters if isinstance(param, BodyParameter)]
        assert len(param_types := {type(param) for param in body_params}) <= 1, f"Can not combine {param_types} in body"
        assert len(param_mimes := {param.body_mime_type for param in body_params}) <= 1, \
            f"Can not combine {param_mimes} in body"
        return sorted(body_params, key=lambda x: x.default is not NoValue)

    @cached_property
    def body_parameter_type(self) -> type[BodyParameter] | None:
        return type(first(self.body_parameters, none_if_empty=True))

    @cached_property
    def non_body_parameters(self) -> list[CallParameter]:
        return [param for param in self.parameters if not isinstance(param, BodyParameter)]

    @cached_property
    def body_content_schema(self) -> type | None:
        if not (body_params := self.body_parameters):
            return None

        if self.body_parameter_type and issubclass(self.body_parameter_type, RawBodyParameter):
            return bytes

        class_: type = type(self.operation_id, tuple(), {
            p.name: field(default=MISSING if p.default is not NoValue else p.default) for p in body_params})
        class_.__annotations__ = {param.name: param.schema for param in body_params}
        # noinspection PyTypeChecker
        return dataclasses.dataclass(class_)

    @cached_property
    def request_body(self) -> RequestBody | None:
        if (body_parameter_type := self.body_parameter_type) and (body_content_schema := self.body_content_schema):
            return RequestBody(Content(body_parameter_type.body_mime_type, body_content_schema))

    @cached_property
    def code_to_answer(self) -> dict[http_tools.HttpStatusCode, type[http_tools.Answer]]:
        result = {self.answer_type.get_class_status_code(): self.answer_type}
        result |= {ans_t.get_class_status_code(): ans_t for ans_t in self.exception_type_to_answer_type.values()}
        if self.unhandled_exception__answer_type is not None:
            unhandled_exception__answer_code = self.unhandled_exception__answer_type.get_class_status_code()
            result[unhandled_exception__answer_code] = self.unhandled_exception__answer_type
        return result

    def as_endpoint(self) -> Endpoint:
        return Endpoint(
            path=self.path,
            method=self.method,
            operation_id=self.operation_id,
            securities=self.securities,
            parameters=[param for param in self.non_body_parameters if isinstance(param, SpecParameter)],
            request_body=self.request_body,
            code_to_answer=self.code_to_answer
        )

    def gen_handler(self) -> Callable[[IncomingRequest], Awaitable[Answer]]:
        # TODO: check argument fullness
        if not self.securities:
            warnings.warn('Endpoints without securities will be prohibited')

        async def handler(request: IncomingRequest) -> Answer:
            kwargs = None
            cache_token = resolver_cache.set({})
            security_key_to_exception = {}
            if self.securities:
                for security in self.securities:
                    security_key = f'{security.scheme.key}[{",".join(security.scopes)}]'
                    try:
                        kwargs = await security.evaluate(request)
                        break
                    except (Unauthorized, Forbidden) as e:
                        if security.scheme.do_log:
                            security_key_to_exception[security_key] = e
            else:
                kwargs = {}
            resolver_cache.reset(cache_token)

            if kwargs is None:
                is_forbidden = any(isinstance(exc, Forbidden) for exc in security_key_to_exception.values())
                status_code = HttpStatusCode.Forbidden if is_forbidden else HttpStatusCode.Unauthorized
                # noinspection PyTypeChecker
                return JsonableAnswer(ErrorDescription(security_key_to_exception, "ErrorGroup"), status_code)

            try:
                kwargs |= {name: await param.get(request) for name, param in self.arg_name_to_param.items()}
                call_result = await acall(self.func(**kwargs))
                answer_factory = self.answer_type if self.answer_factory is None else self.answer_factory
                return answer_factory(call_result)
            except Exception as e:
                for exception_type, answer_type in self.exception_type_to_answer_type.items():
                    if isinstance(e, exception_type):
                        return answer_type(e)

                if self.unhandled_exception__answer_type is not None:
                    traceback.print_exc()
                    logger.exception(e)
                    return self.unhandled_exception__answer_type(e)

                raise

        return handler


class OpenApiServer:
    @dataclass
    class Context(OpenApiWrapper.Context):
        http_server: HttpServer

    @dataclass
    class Config(OpenApiWrapper.Config):
        write_spec_to_file: str = ""  # empty means skip
        spec_http_path: str = '/open_api_specification/get'  # empty means disable

    def __init__(
            self, config: Config, context: Context,
            wrapper_factory: Callable[[Config, Context], OpenApiWrapper] = OpenApiWrapper
    ):
        self.context = context
        self.config = config
        self.wrapper = wrapper_factory(config, context)
        if path := self.config.spec_http_path:
            self.context.http_server.register_handler(path, self.get_specification)

    async def get_specification(self, _: IncomingRequest) -> FileAnswer:
        content = custom_dumps(self.wrapper.get_spec_dict(), indent=2)
        return FileAnswer(content, file_name='openapi_spec.json', content_type=ContentType.Json)

    def _write_spec_to_file(self):
        if self.config.write_spec_to_file:
            spec = self.wrapper.get_spec_dict()
            json_res = json.dumps(spec, indent=2)
            with open(self.config.write_spec_to_file, "w") as f:
                f.write(json_res)

    def register_endpoint(self, rpc_endpoint: RpcEndpoint) -> None:
        self.context.http_server.register_handler(rpc_endpoint.path, rpc_endpoint.gen_handler(), [rpc_endpoint.method])
        self.wrapper.register_endpoint(rpc_endpoint.as_endpoint())
        self._write_spec_to_file()
